package cn.qqck.kotlin.tools

fun ByteArray.string(): String = this.toString(java.nio.charset.Charset.defaultCharset())

/**
 * 压缩当前字节数组，返回压缩后的字节数组。
 *
 * @param level 压缩级别，范围为 `0`（不压缩）到 `9`（最佳压缩）。默认为 `java.util.zip.Deflater.BEST_COMPRESSION`。
 * @return 一个新的字节数组，包含压缩后的数据。如果压缩失败或数据为空，则可能返回空数组。
 */
fun ByteArray.compress(level: Int = java.util.zip.Deflater.BEST_COMPRESSION): ByteArray {
    runDefer {
        val j_infl = java.util.zip.Deflater(level)
        defer { j_infl.end() }
        j_infl.setInput(this@compress)
        val j_bufs = java.io.ByteArrayOutputStream()
        defer { j_bufs.close() }
        val j_out = ByteArray(4096)
        while (!j_infl.finished()) {
            val j_size = j_infl.deflate(j_out)
            if (j_size == 0) break
            j_bufs.write(j_out, 0, j_size)
        }
        return j_bufs.toByteArray()
    }
}

/**
 * 解压缩当前字节数组，返回解压后的数据。
 * 使用 `java.util.zip.Inflater` 进行解压操作。
 *
 * @return 解压后的字节数组。如果解压失败，可能抛出相关异常。
 */
fun ByteArray.uncompress(): ByteArray {
    runDefer {
        val j_infl = java.util.zip.Inflater()
        defer { j_infl.end() }
        j_infl.setInput(this@uncompress)
        val j_bufs = java.io.ByteArrayOutputStream()
        defer { j_bufs.close() }
        val j_out = ByteArray(4096)
        while (!j_infl.finished()) {
            val j_size = j_infl.inflate(j_out)
            if (j_size == 0) break
            j_bufs.write(j_out, 0, j_size)
        }
        return j_bufs.toByteArray()
    }
}

/**
 * 从当前的字节数组复制指定范围的字节，并返回一个新的字节数组。
 *
 * @param start 起始位置的索引，从该索引开始复制内容。如果起始位置超出当前数组范围，则返回的数组会包含默认值。
 * @param size  目标字节数组的大小，即返回的字节数组长度。
 * @return 包含复制内容的新字节数组，长度为指定的 `size`。
 */
fun ByteArray.copyOf(start: Int, size: Int): ByteArray {
    val j_bufs = ByteArray(size)
    if (start < this.size) System.arraycopy(this, start, j_bufs, 0, kotlin.math.min(this.size, size))
    return j_bufs
}

/**
 * 对当前字节数组与指定的密钥字节数组进行按位异或运算。
 * 如果当前字节数组或密钥字节数组为空，则返回一个空的字节数组。
 *
 * @param key 用于异或运算的密钥字节数组。如果密钥长度小于当前字节数组长度，会循环使用密钥。
 * @return 返回异或运算后的字节数组，长度与当前字节数组一致。
 */
fun ByteArray.xor(key: ByteArray): ByteArray {
    if (this.isEmpty() || key.isEmpty()) return ByteArray(0)
    val j_keysize = key.size
    return ByteArray(size) { j_i -> (this[j_i].toInt() xor key[j_i % j_keysize].toInt()).toByte() }
}

/**
 * 将字节数组的指定位置转换为短整型（Short）。
 *
 * @param start 开始读取的字节数组索引位置，默认值为0。
 * @param big 指定字节序是否为大端字节序（Big-endian）。若为false，则为小端字节序（Little-endian）。默认值为false。
 * @return 从字节数组中转换得到的短整型（Short）。若字节数组长度不足以读取，则返回0。
 */
fun ByteArray.toShort(start: Int = 0, big: Boolean = false): Short {
    return if (this.size < start + 2) 0 else if (big) {
        ((this[start].toUByte().toInt() shl 8) or
                this[start + 1].toUByte().toInt()).toShort()
    } else {
        (this[start].toUByte().toInt() or
                (this[start + 1].toUByte().toInt() shl 8)).toShort()
    }
}

/**
 * 将ByteArray中的指定起始位置的两个字节转换为UShort。
 *
 * @param start 起始位置的索引，默认为0。
 * @param big 表示是否按大端字节序读取，true为大端字节序，false为小端字节序，默认为false。
 * @return 转换后的UShort值。如果ByteArray的长度不足以读取两个字节，则返回0U。
 */
fun ByteArray.toUShort(start: Int = 0, big: Boolean = false): UShort {
    return if (this.size < start + 2) 0U else if (big) {
        ((this[start].toUByte().toInt() shl 8) or
                this[start + 1].toUByte().toInt()).toUShort()
    } else {
        (this[start].toUByte().toInt() or
                (this[start + 1].toUByte().toInt() shl 8)).toUShort()
    }
}

/**
 * 将字节数组的指定部分转换为整数。
 *
 * @param start 起始索引，用于指定从字节数组的哪个位置开始读取，默认为0。
 * @param big 如果为true，则使用大端模式（字节序从高到低），否则使用小端模式（字节序从低到高），默认为false。
 * @return 转换后的整数值。如果字节数组的长度不足以读取4个字节，将返回0。
 */
fun ByteArray.toInt(start: Int = 0, big: Boolean = false): Int {
    return if (this.size < start + 4) 0 else if (big) {
        (this[start].toUByte().toInt() shl 24) or
                (this[start + 1].toUByte().toInt() shl 16) or
                (this[start + 2].toUByte().toInt() shl 8) or
                this[start + 3].toUByte().toInt()
    } else {
        this[start].toUByte().toInt() or
                (this[start + 1].toUByte().toInt() shl 8) or
                (this[start + 2].toUByte().toInt() shl 16) or
                (this[start + 3].toUByte().toInt() shl 24)
    }
}

/**
 * 将字节数组中的一部分转换为无符号整型 (UInt)。
 *
 * @param start 起始索引，默认值为0，用于指定从字节数组的哪一部分开始读取数据。
 * @param big 是否采用大端模式。true 表示采用大端模式，false 表示采用小端模式。
 * @return 转换得到的无符号整型。如果字节数组长度不足以读取4个字节，则返回0U。
 */
fun ByteArray.toUInt(start: Int = 0, big: Boolean = false): UInt {
    return if (this.size < start + 4) 0U else if (big) {
        (this[start].toUByte().toUInt() shl 24) or
                (this[start + 1].toUByte().toUInt() shl 16) or
                (this[start + 2].toUByte().toUInt() shl 8) or
                this[start + 3].toUByte().toUInt()
    } else {
        this[start].toUByte().toUInt() or
                (this[start + 1].toUByte().toUInt() shl 8) or
                (this[start + 2].toUByte().toUInt() shl 16) or
                (this[start + 3].toUByte().toUInt() shl 24)
    }
}

/**
 * 将当前字节数组转换为 `Long` 值。
 *
 * @param start 起始位置，表示从数组的哪个索引开始读取数据。默认为 0。
 * @param big 是否使用大端序进行解析。如果为 `true`，则按照大端序解析；否则按照小端序解析。默认为 `false`。
 * @return 转换后的 `Long` 值。如果字节数组长度不足以解析 `Long`（即从 `start` 开始不足 8 字节），则返回 0。
 */
fun ByteArray.toLong(start: Int = 0, big: Boolean = false): Long {
    return if (this.size < start + 8) 0 else if (big) {
        (this[start].toUByte().toLong() shl 56) or
                (this[start + 1].toUByte().toLong() shl 48) or
                (this[start + 2].toUByte().toLong() shl 40) or
                (this[start + 3].toUByte().toLong() shl 32) or
                (this[start + 4].toUByte().toLong() shl 24) or
                (this[start + 5].toUByte().toLong() shl 16) or
                (this[start + 6].toUByte().toLong() shl 8) or
                this[start + 7].toUByte().toLong()
    } else {
        this[start].toUByte().toLong() or
                (this[start + 1].toUByte().toLong() shl 8) or
                (this[start + 2].toUByte().toLong() shl 16) or
                (this[start + 3].toUByte().toLong() shl 24) or
                (this[start + 4].toUByte().toLong() shl 32) or
                (this[start + 5].toUByte().toLong() shl 40) or
                (this[start + 6].toUByte().toLong() shl 48) or
                (this[start + 7].toUByte().toLong() shl 56)
    }
}

/**
 * 将字节数组从指定起始位置转为无符号长整型值 (ULong)。
 *
 * @param start 起始位置索引，默认为0。
 * @param big 是否按大端字节序处理，默认为false，小端字节序。
 * @return 转换后的无符号长整型值。如果字节数组长度不足以支持转换，则返回0U。
 */
fun ByteArray.toULong(start: Int = 0, big: Boolean = false): ULong {
    return if (this.size < start + 8) 0U else if (big) {
        (this[start].toUByte().toULong() shl 56) or
                (this[start + 1].toUByte().toULong() shl 48) or
                (this[start + 2].toUByte().toULong() shl 40) or
                (this[start + 3].toUByte().toULong() shl 32) or
                (this[start + 4].toUByte().toULong() shl 24) or
                (this[start + 5].toUByte().toULong() shl 16) or
                (this[start + 6].toUByte().toULong() shl 8) or
                this[start + 7].toUByte().toULong()
    } else {
        this[start].toUByte().toULong() or
                (this[start + 1].toUByte().toULong() shl 8) or
                (this[start + 2].toUByte().toULong() shl 16) or
                (this[start + 3].toUByte().toULong() shl 24) or
                (this[start + 4].toUByte().toULong() shl 32) or
                (this[start + 5].toUByte().toULong() shl 40) or
                (this[start + 6].toUByte().toULong() shl 48) or
                (this[start + 7].toUByte().toULong() shl 56)
    }
}
