package cn.qqck.kotlin.tools

class RunTimeScope {
    private var stop = false
    fun isstop(): Boolean = this.stop
    fun stop() {
        this.stop = true
    }
}

/**
 * 周期性执行块代码直到超时或满足停止条件。
 *
 * @param mill 超时时间，单位为毫秒。
 * @param sleep 每次循环间的休眠时间，默认为10毫秒。
 * @param block 使用`RunTimeScope`上下文作为接收者的代码块。
 * @return 返回代码块的最终执行结果。
 */
inline fun <T> runTime(mill: Long, sleep: Long = 10, block: RunTimeScope.() -> T): T {
    val j_timeout = System.currentTimeMillis() + mill
    val j_runtime = RunTimeScope()
    while (true) {
        val j_endTime = j_timeout - System.currentTimeMillis()
        val j_ret = j_runtime.block()
        if (j_endTime < 1 || j_runtime.isstop()) return j_ret
        Thread.sleep(sleep)
    }
}

class DeferScope {
    private val defers = ArrayDeque<() -> Unit>()

    fun defer(block: () -> Unit) {
        this.defers.addLast(block)
    }

    fun release() {
        while (this.defers.isNotEmpty()) this.defers.removeLast()()
    }
}

/**
 * 执行一个带有延迟执行块的作用域代码，并在完成后释放所有延迟任务。
 *
 * @param block 需要在延迟作用域中执行的代码块。该代码块在提供的 [DeferScope] 上调用，
 *              允许使用 `defer` 方法添加延迟执行的任务。
 * @return 返回代码块执行后的结果。
 */
inline fun <T> runDefer(block: DeferScope.() -> T): T {
    val j_defer = DeferScope()
    val j_ret = try {
        j_defer.block()
    } finally {
        j_defer.release()
    }
    return j_ret
}
