package cn.qqck.kotlin.tools

import java.io.File
import java.nio.ByteBuffer
import java.nio.channels.FileChannel
import java.nio.file.Paths
import java.nio.file.StandardOpenOption

/**
 * 将字符串数据写入指定路径的文件。
 *
 * @param data 待写入的字符串数据。
 * @param dir 文件保存的目录路径，可选参数，默认为 `null`（当前路径下操作）。
 * @param append 是否以追加的方式写入文件，默认为 `false`（覆盖模式写入）。
 * @return 如果写入成功返回 `true`；否则返回 `false`。
 */
fun String.fset(data: String, dir: String? = null, append: Boolean = false): Boolean = this.fset(data.toByteArray(), dir, append)

/**
 * 将二进制数据写入指定路径的文件。
 *
 * @param data 待写入的字节数组。
 * @param dir  文件保存的目录路径，可选参数，默认为 `null`（当前路径下操作）。
 * @param append 是否以追加的方式写入文件，默认为 `false`（覆盖模式写入）。
 * @return 如果写入成功返回 `true`；否则返回 `false`。
 */
fun String.fset(data: ByteArray, dir: String? = null, append: Boolean = false): Boolean {
    val j_name = if (dir == null) Paths.get(this) else Paths.get(dir, this)
    do {
        try {
            val j_c = FileChannel.open(j_name, StandardOpenOption.CREATE, if (append) StandardOpenOption.APPEND else StandardOpenOption.WRITE)
            do {
                try {
                    val j_lock = j_c.lock()
                    try {
                        j_c.write(ByteBuffer.wrap(data))
                        if (!append) j_c.truncate(j_c.position())
                        return true
                    } catch (_: Exception) {
                        return false
                    } finally {
                        j_lock.close()
                        j_c.close()
                    }
                } catch (_: java.nio.channels.OverlappingFileLockException) {
                    Thread.sleep(1)
                } catch (_: Exception) {
                    j_c.close()
                    return false
                }
            } while (true)
        } catch (e: java.nio.file.NoSuchFileException) {
            if (dir == null || !File(dir).mkdirs()) return false
        } catch (_: Exception) {
            return false
        }
    } while (true)
}

/**
 * 从当前字符串表示的文件路径中读取内容并返回其字节数组。
 * 如果文件使用过程中出现异常、大小超出限制或读取失败，则返回 null。
 *
 * @return 文件内容的字节数组，若读取失败则返回 null。
 */
fun String.fget(): ByteArray? {
    try {
        val j_c = FileChannel.open(Paths.get(this), StandardOpenOption.WRITE, StandardOpenOption.READ)
        do {
            try {
                val j_lock = j_c.lock()
                try {
                    val j_size = j_c.size()
                    if (j_size > Int.MAX_VALUE) return null
                    val j_buf = ByteBuffer.allocate(j_size.toInt())
                    if (j_c.read(j_buf) != j_size.toInt()) return null
                    return j_buf.array()
                } catch (_: Exception) {
                    return null
                } finally {
                    j_lock.close()
                    j_c.close()
                }
            } catch (_: java.nio.channels.OverlappingFileLockException) {
                Thread.sleep(1)
            } catch (e: Exception) {
                j_c.close()
                return null
            }
        } while (true)
    } catch (_: Exception) {
    }
    return null
}
