package cn.qqck.kotlin.tools

import kotlin.concurrent.withLock

class Proxy {
    enum class Type {
        Direct,
        Http,
        Socks5,
    }

    data class Info(
        var addr: String = "",
        var port: Int = 0,
        var user: String = "",
        var pass: String = "",
        var type: Type = Type.Http,
        var timeout: Long = 0,
    ) {

        /**
         * 判断代理是否有效且时间是否未到期
         */
        fun isValid(): Boolean {
            if (this.addr == "" || this.port < 1 || this.port > 65535) return false
            return this.timeout == -1L || System.currentTimeMillis() < this.timeout
        }

        /**
         * 代理是否需要授权
         */
        fun authorization(): Boolean {
            return this.user != "" && this.pass != ""
        }

        /**
         * 编码为"Type://user:pass@host:port"，例如："Http://user:pass@127.0.0.1:808"、"socks5://user:pass@127.0.0.1:808"
         */
        override fun toString(): String {
            if (this.addr == "") return ""
            val j_buf = StringBuilder(64)
            j_buf.append(
                when (this.type) {
                    Type.Direct -> "direct://"
                    Type.Http -> "http://"
                    Type.Socks5 -> "socks5://"
                }
            )
            if (this.authorization()) {
                j_buf.append(this.user)
                j_buf.append(":")
                j_buf.append(this.pass)
                j_buf.append("@")
            }
            j_buf.append(this.addr)
            j_buf.append(":")
            j_buf.append(this.port)
            return j_buf.toString()
        }
    }

    /**
     * 代理池
     */
    class Pool {
        private val lock = java.util.concurrent.locks.ReentrantLock()

        /**
         * 提取链接或直连地址
         */
        private var url: String = ""

        /**
         * 提取链接或直连地址
         */
        fun url(): String {
            lock.withLock {
                return this.url
            }
        }

        /**
         * 提取链接或直连地址
         */
        fun url(value: String): Pool {
            lock.withLock { this.url = value }
            return this
        }


        /**
         * 提取链接是否为隧道直连地址
         */
        private var url_tunnel: Boolean = false

        /**
         * 提取链接是否为隧道直连地址
         */
        fun url_tunnel(): Boolean {
            lock.withLock { return this.url_tunnel }
        }

        /**
         * 提取链接是否为隧道直连地址
         */
        fun url_tunnel(value: Boolean): Pool {
            lock.withLock { this.url_tunnel = value }
            return this
        }


        /**
         * 请求 Url 的最少间隔，默认不限制，单位：毫秒
         */
        private var url_space: Long = 0

        /**
         * 请求 Url 的最少间隔，默认不限制，单位：毫秒
         */
        fun url_space(): Long {
            lock.withLock { return this.url_space }
        }

        /**
         * 请求 Url 的最少间隔，默认不限制，单位：毫秒
         */
        fun url_space(value: Long): Pool {
            lock.withLock { this.url_space = value }
            return this
        }


        /**
         * 代理类型，默认[Type.Http]
         */
        private var type: Type = Type.Http

        /**
         * 代理类型，默认[Type.Http]
         */
        fun type(): Type {
            lock.withLock { return this.type }
        }

        /**
         * 代理类型，默认[Type.Http]
         */
        fun type(value: Type): Pool {
            lock.withLock { this.type = value }
            return this
        }


        /**
         * 认证账号，如果代理需要身份认证的情况下填写
         */
        private var user: String = ""

        /**
         * 认证账号，如果代理需要身份认证的情况下填写
         */
        fun user(): String {
            lock.withLock { return this.user }
        }

        /**
         * 认证账号，如果代理需要身份认证的情况下填写
         */
        fun user(value: String): Pool {
            lock.withLock { this.user = value }
            return this
        }


        /**
         * 认证密码，如果代理需要身份认证的情况下填写
         */
        private var pass: String = ""

        /**
         * 认证密码，如果代理需要身份认证的情况下填写
         */
        fun pass(): String {
            lock.withLock { return this.pass }
        }

        /**
         * 认证密码，如果代理需要身份认证的情况下填写
         */
        fun pass(value: String): Pool {
            lock.withLock { this.pass = value }
            return this
        }


        /**
         * 代理有效时长，默认30秒，单位：毫秒
         */
        private var valid_time: Long = 1000 * 30

        /**
         * 代理有效时长，默认30秒，单位：毫秒
         */
        fun valid_time(): Long {
            lock.withLock { return this.valid_time }
        }

        /**
         * 代理有效时长，默认30秒，单位：毫秒
         */
        fun valid_time(value: Long): Pool {
            lock.withLock { this.valid_time = value }
            return this
        }


        /**
         * 代理池内有效时间低于该值就视为无效代理，默认10秒，单位：毫秒
         */
        private var release_time: Long = 1000 * 10

        /**
         * 代理池内有效时间低于该值就视为无效代理，默认10秒，单位：毫秒
         */
        fun release_time(): Long {
            lock.withLock { return this.release_time }
        }

        /**
         * 代理池内有效时间低于该值就视为无效代理，默认为10秒，单位：毫秒
         */
        fun release_time(value: Long): Pool {
            lock.withLock { this.release_time = value }
            return this
        }


        // -------------------------------------------------------------------------


        /**
         * 代理池列表
         */
        private var get_list = ArrayDeque<Info>()

        /**
         * 当前代理池剩余[Info]数量
         */
        fun size(): Int {
            return lock.withLock { this.get_list.size }
        }

        /**
         * 最后请求 Url 的时间，单位：毫秒
         */
        private var get_request_time: Long = 0

        /**
         * 已使用计数
         */
        private var get_i: Int = 0

        /**
         * 单个 IP 最多使用次数，默认为：1
         */
        private var get_max: Int = 1

        /**
         * 单个 IP 最多使用次数，默认为：1
         */
        fun get_max(): Int {
            lock.withLock { return this.get_max }
        }

        /**
         * 单个 IP 最多使用次数，默认为：1
         */
        fun get_max(value: Int): Pool {
            lock.withLock { this.get_max = if (value < 1) 1 else value }
            return this
        }

        /**
         * 已提取计数
         */
        private var get_count = 0L

        /**
         * 已提取计数
         */
        fun get_count(): Long {
            lock.withLock { return this.get_count }
        }

        /**
         * 当前代理
         */
        private var get_info: Info = Info()


        /**
         * 重置代理池缓存信息
         */
        fun reset(): Pool {
            lock.withLock {
                this.get_list.clear()
                this.get_request_time = 0
                this.get_i = 0
                this.get_count = 0
                this.get_info = Info()
            }
            return this
        }

        /**
         * 获取一个有效的代理信息。
         * 方法会根据当前配置及代理池状态，返回一个可用的 `Info` 对象。
         * 如果当前代理池为空或无可用代理信息，可能返回 `null`。
         *
         * @return 一个有效的 `Info` 对象，若无可用代理则返回 `null`。
         */
        fun get(): Info? {
            lock.withLock {
                if (this.url_tunnel) {
                    val j_adp = this.url.toAddrPort() ?: return null
                    this.get_count++
                    return Info(j_adp.addr, j_adp.port, this.user, this.pass, this.type, System.currentTimeMillis() + this.valid_time - this.release_time)
                }
                if (!this.get_info.isValid() || this.get_i >= this.get_max) {
                    if (this.get_list.isEmpty()) this.get_url()
                    if (this.get_list.isEmpty()) return null
                    while (this.get_list.isNotEmpty()) {
                        this.get_info = this.get_list.removeAt(0)
                        if (this.get_info.isValid()) {
                            this.get_i = 1
                            this.get_count++
                            return this.get_info
                        }
                    }
                    return null
                }
                this.get_i++
                return this.get_info
            }
        }

        private val url_item_sps = arrayOf(" ", "|", ":", "/")
        private fun get_url() {
            if (this.url == "") return
            var j_time = System.currentTimeMillis()
            if (j_time - this.get_request_time < this.url_space) return
            Okhttp.call(this.url, Okhttp.Option().redirects(true))?.also {
                if (it.code != 200) return
                val j_response = it.body?.string() ?: return
                this.get_request_time = j_time
                j_time = System.currentTimeMillis() + this.valid_time - this.release_time
                (if (j_response.contains("\r\n")) {
                    j_response.split("\r\n")
                } else if (j_response.contains("\\r\\n")) {
                    j_response.split("\\r\\n")
                } else if (j_response.contains("\r")) {
                    j_response.split("\r")
                } else if (j_response.contains("\\r")) {
                    j_response.split("\\r")
                } else if (j_response.contains("\n")) {
                    j_response.split("\n")
                } else if (j_response.contains("\\n")) {
                    j_response.split("\\n")
                } else if (j_response.contains("\\t")) {
                    j_response.split("\\t")
                } else if (j_response.contains("\t")) {
                    j_response.split("\t")
                } else if (j_response.contains("<br/>")) {
                    j_response.split("<br/>")
                } else if (j_response.contains("    ")) {
                    j_response.split("    ")
                } else if (j_response.contains("   ")) {
                    j_response.split("   ")
                } else if (j_response.contains("  ")) {
                    j_response.split("  ")
                } else {
                    listOf(j_response)
                }).forEach { j_item ->
                    if (j_item.isEmpty()) return@forEach
                    val j_info = Info("", 0, this.user, this.pass, this.type, j_time)
                    for (j_sp in this.url_item_sps) if (j_item.indexOf(j_sp) > -1) {
                        val j_items = j_item.split(j_sp)
                        if (j_items.size == 3) {
                            val j_adp = j_items[0].toAddrPort() ?: return@forEach
                            j_info.addr = j_adp.addr
                            j_info.port = j_adp.port
                            j_info.user = j_items[1]
                            j_info.pass = j_items[2]
                            this.get_list.add(j_info)
                            return@forEach
                        } else if (j_items.size == 4) {
                            j_info.addr = j_items[0]
                            j_info.port = j_items[1].toIntOrNull() ?: return@forEach
                            j_info.user = j_items[2]
                            j_info.pass = j_items[3]
                            this.get_list.add(j_info)
                            return@forEach
                        }
                    }
                    val j_adp = j_item.toAddrPort() ?: return@forEach
                    j_info.addr = j_adp.addr
                    j_info.port = j_adp.port
                    this.get_list.add(j_info)
                }
            }
        }
    }
}
