package cn.qqck.kotlin.tools

private val units = arrayOf("B", "KB", "MB", "GB", "TB")

/**
 * 将当前 Short 类型值转换为单位表示的字符串形式。
 *
 * @return 表示该值的单位字符串，例如以 B, KB, MB 等单位表示的结果。
 */
fun Short.toUnits(): String = this.toULong().toUnits()

/**
 * 将当前无符号短整型（UShort）数值转换为便于人类阅读的字符串表示，通常用于表示存储单位（如 B、KB、MB 等）。
 *
 * @return 转换后的字符串表示形式，示例格式为 "12.34 KB"。如果值为 0，则返回 "0 B"。
 */
fun UShort.toUnits(): String = this.toULong().toUnits()

/**
 * 将当前整型值转换为以适当单位表示的字符串。
 *
 * @return 转换后的字符串，表示适当单位的人类可读格式，例如B（字节）、KB（千字节）、MB（兆字节）等。
 */
fun Int.toUnits(): String = this.toULong().toUnits()

/**
 * 将当前无符号整型数值转换为人类可读的单位格式字符串。
 *
 * 例如，将内存大小或文件大小等数值表示为带单位的字符串形式（如KB, MB, GB等）。
 *
 * @return 转换后的带单位格式的字符串。
 */
fun UInt.toUnits(): String = this.toULong().toUnits()

/**
 * 将当前 Long 类型的数值转换为以适当单位表示的字符串。
 *
 * @return 转换后的字符串形式，表示大小的数值单位（例如 "1.00 KB", "1.00 MB" 等）。
 */
fun Long.toUnits(): String = this.toULong().toUnits()

/**
 * 将当前的无符号长整型值转换为以单位（如B、KB、MB等）表示的字符串。
 *
 * 该方法会根据数值的大小，自动选择合适的单位，将数值转换为更易读的形式。
 *
 * @return 表示转换后的带单位的字符串，例如 "12.34 KB" 或 "1.00 MB"。
 */
fun ULong.toUnits(): String {
    if (this < 0u) return "0 B"
    var j_size = this.toDouble()
    var j_i = 0
    while (j_size >= 1024 && j_i < units.size - 1) {
        j_size /= 1024
        j_i++
    }
    return String.format("%.2f %s", j_size, units[j_i])
}
